# bash base library version 0.6.8
# Copyright (C)  2015-2024 Heinz Deinhart
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This is a bash library, it has to be sourced not called.
#
# bash-base provides a number of functions and a base setup of settings
# to support clean coding style in bash.
# shellcheck shell=bash
#
# Internal global veriables are prefixed with _BASH_BASE_, internal
# function with _bash_base_

[ "$0" != "${BASH_SOURCE[0]}" ] || {
	echo "$0: this script must be sourced, not called" >&2
	exit 1
}

#@@ environment variables
#   BASH_BASE_ASSERT_MODE - can be 0, 1 or 2
#     0 .. disable asserts, makes bash-base faster but less type safe
#     1 .. enable asserts and emit log messages when one fails
#     2... enable asserts and fail with bb_fatal if one fails

#@@ error messages

#@ error-codes (bash-base)
#   1 .. unkown internal error
#   2 .. illegal number of arguments in function call
#   3 .. illegal variable name
#   4 .. illegal function arguments
#   5 .. error sourcing file
#   6 .. no such function
#   7 .. variable specified as argument undefined
#   8 .. no such file
#   9 .. no access rights
#  10 .. value not found
#  11 .. incorrect variable type
#  12 .. variable specified as argument is defined but unset [[ ! -v  ]]
#  9x .. errors local to each function

#@ bb_error_string <ret-vn> <error-code>
# errors: 2, 3
function bb_error_string() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _err_code="$2"

    bb_vn_validate "$_ret_vn" "bb_error_string:ret-vn" || return 3

    local _msg
    case "$_err_code" in
	0) _msg="no error";;
	1) _msg="unkown internal error";;
	2) _msg="illegal number of arguments in function call";;
	3) _msg="illegal variable name";;
	4) _msg="illegal function arguments";;
	5) _msg="error sourcing file";;
	6) _msg="no such function";;
	7) _msg="variable specified as argument undefined";;
	8) _msg="no such file";;
	9) _msg="no access rights";;
	10) _msg="value not found";;
	11) _msg="incorrect variable type";;
	12) _msg="variable specified as argument is defined but unset";;
	90|91|92|93|94|95|96|97|98|99)
	    _msg="function local error $_err_code";;
	*) _msg="bb_error_string: unknown error code $_err_code";;
    esac
    printf -v "$_ret_vn" "%s" "$_msg"
}

#@@ type related functions

#@ bb_var_type <ret-basetype-vn> <ret-mod-vn> <ret-set-vn> <var-vn>
# returns the type of a var where basetype can be: str, array or dict
# mod returns the modifiers as returned by declare -p (see bash docu)
# ret-basetype-vn and ret-mod-vn can be empty and then will not be used
# note: this works on defined vars, whether they are set or unset
# errors: 2, 7
function bb_var_type() {
    (( $# == 4 )) || {
	bb_ierr "called with $# instead of 4 arg"
	return 2
    }

    local _ret_type_vn="$1"
    local _ret_mod_vn="$2"
    local _ret_set_vn="$3"
    local _var_vn="$4"

    local _line _declare_ouput _signature _type _mod
    local -i _set=0

    _line="$(declare -p "$_var_vn" 2> /dev/null)" || return 7
    _declare_output="${_line#declare -}"
    _signature="${_declare_output%% *}"

    # let check if variable is set or unset
    if [[ "$_declare_output" =~ = ]]
    then # set
	_set=1
    fi

    case "$_signature" in
	*a*) _type="array";;
	*A*) _type="dict";;
	*n*) _type="nameref";;
	*) _type="string"
    esac

    [ -z "$_ret_type_vn" ] || printf -v "$_ret_type_vn" "%s" "$_type"
    [ -z "$_ret_mod_vn" ] || printf -v "$_ret_mod_vn" "%s" "$_signature"
    [ -z "$_ret_set_vn" ] || printf -v "$_ret_set_vn" "%i" "$_set"
}

#@ bb_vn_validate <var-name> [<extra-info>]
# validates variable names
# valid variable names match the following regex: ^[[:alpha:]][[:alnum:]_]*$
# if extra-info is non empty, a log msg with level warning will be emitted
# when var-name is invalid
# errors: 2, 3
function bb_vn_validate() {
    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _vn="$1"
    local _xinfo="${2-}"

    [[ "$_vn" =~ ^[[:alpha:]][[:alnum:]_]*$ ]] || {
	[ -z "$_xinfo" ] || bb_msg warning \
	    "${FUNCNAME[1]-?}:$_xinfo: illegal variable name: $_vn"
	return 3
    }
}

#@ bb_uint_validate <uint> [..]
# validates if one or more arguments are unsigned integers
# errors: 2, 3
function bb_uint_validate() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    while (( $# > 0 ))
    do
	[[ "$1" =~ ^[[:digit:]]+$ ]] || return 3
	shift
    done
}

#@ bb_int_validate <int> [..]
# validates if one or more arguments are integers
# errors: 2, 3
function bb_int_validate() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    while (( $# > 0 ))
    do
	[[ "$1" =~ ^[[:digit:]-][[:digit:]]*$ ]] || return 3
	shift
    done
}

#@ bb_fun_is_defined <function-name>
# checks if a functino with name function-name is defined
# errors: 2, 6
function bb_fun_is_defined() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    declare -f "$1" >/dev/null || return 6
}

#@@ executable related funs

#@ bb_exec_is_callable <exec-name>
# checks if exec-name is a callable function or executable in PATH
# errors: 2, 8
function bb_exec_is_callable() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _ex="$1"
    local _xinfo="${2-}"

    if declare -f "$_ex" >/dev/null
    then
	return 0 # function
    fi

    if [[ "$_ex" =~ ^((\.?\./)|/) ]]
    then
	# absolute or relative path included
	if [[ -x "$_ex" ]]
	then
	    return 0
	fi
    else
	# executable might be in PATH, check
	local _found_ex
	_found_ex="$(which "$_ex" 2>/dev/null || true)"

	if [[ "$_found_ex" ]]
	then
	    return 0
	fi
    fi

    return 8
}

#@@ string related functions

#@ bb_string_is_defined <vn>
# check if variable is defined and of type string
# errors: 2, 7, 11
function bb_string_is_defined() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local bb_string_is_defined__type
    bb_var_type bb_string_is_defined__type "" "" "$1" || return 7
    [ "$bb_string_is_defined__type" == "string" ] || return 11
}

#@ bb_string_is_set <vn>
# check if variable is string and set (has been assigned a value)
# errors: 2, 7, 11, 12
function bb_string_is_set() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local bb_string_is_set__type bb_string_is_set__set
    bb_var_type bb_string_is_set__type "" bb_string_is_set__set "$1" || \
	return 7
    [ "$bb_string_is_set__type" == "string" ] || return 11
    [ "$bb_string_is_set__set" == "1" ] || return 12
}

#@ bb_string_compare <string1> <string2>
# compare if two strings are equal, return 0 if they are
# errors: 2, 91 (not equal)
function bb_string_compare() {
   (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

   [[ "$1" == "$2" ]] || return 91
}

#@ bb_string_quote <ret-vn> <string> [<quotation-string>]
# quote a string: ret-vn := <quotation-string><string><quotation-string>
# if no quotation-string is specified, a quotation mark is used
function bb_string_quote() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "called with $# instead 2 or 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _src="$2"
    local _xarg="${3-\"}"

    bb_assert_string_is_defined "$_ret_vn"

    printf -v "$_ret_vn" "%s%s%s" "$_xarg" "$_src" "$_xarg"
}

#@ bb_string_trim <ret-vn> <string> [left|right|both]
# removes whitespaces from string sides, if no side is specified both is used
# errors: 2, 3, 4
function bb_string_trim() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "called with $# instead of 2 or 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _src="$2"
    local _mode="${3-both}"

    bb_string_trim_custom "$_ret_vn" "$_src" '[:space:]' "$_mode"
}

#@ bb_string_trim_unicode <ret-vn> <string> [left|right|both]
# removes whitespaces from string sides, if no side is specified both is used
# errors: 2, 3, 4
function bb_string_trim_unicode() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "called with $# instead of 2 or 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _src="$2"
    local _mode="${3-both}"

    local _unicode_spaces
    # note: bash4.2 does not successfully expand \uC2A0, bash4.4 does
    printf -v _unicode_spaces "%s\xC2\xA0\u1680\u180E\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200A\u200B\u202F\u205F\u3000\uFEFF" \
	'[:space:]'

    bb_string_trim_custom "$_ret_vn" "$_src" "$_unicode_spaces"	"$_mode"
}
#@ bb_string_trim_custom <ret-vn> <string> <extra-chars> [left|right|both]
# removes whitespaces from string sides, if no side is specified both is used
# errors: 2, 3, 4
function bb_string_trim_custom() {
    (( $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    local _ret_vn="$1"
    local _src="$2"
    local _xc="$3"
    local _mode="${4-both}"

    bb_assert_string_is_defined "$_ret_vn"

    local _regex
    printf -v _regex "%s%s%s%s%s%s%s" '^([' "$_xc" \
	']*)(|(.*[^' "$_xc" ']))([' "$_xc" ']*)$'

    if [[ "$_src" =~ $_regex ]]
    then
	case "$_mode" in
	    both) printf -v "$_ret_vn" "%s" "${BASH_REMATCH[2]}";;
	    left) printf -v "$_ret_vn" "%s%s" \
		"${BASH_REMATCH[2]}" "${BASH_REMATCH[4]}";;
	    right) printf -v "$_ret_vn" "%s%s" \
		"${BASH_REMATCH[1]}" "${BASH_REMATCH[2]}";;
	    *)
		bb_ierr "illegal mode: $_mode"
		return 4
	esac
    else
	bb_ierr "regex did not match but should"
    fi
}

#@ bb_string_split <ret-array-vn> <string> <separator> [<limit>]
# splits string by separator string
# if limit is specified then it splits into at most limit parts starting
# from the left side, limit must be a number > 0
# errors: 2, 3, 4
function bb_string_split() {
    ((  $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    local _ret_array_vn="$1"
    local _string="$2"
    local _separator="$3"
    local _limit="${4-}"

    bb_assert_array_is_defined "$_ret_array_vn"

    [ -z "$_limit" ] || {
	bb_uint_validate "$_limit" || return 4
	(( _limit > 0 )) || return 4
    }

    local -a bb_string_split__result=()
    local _left _right _old_right
    local _count=1

    _right="$_string"
    while (( ${#_right} > 0 ))
    do
	[ -z "$_limit" ] || {
	    if (( _count >= _limit ))
	    then
		# limit was hit
		bb_string_split__result+=("$_right")
		break
	    else
		_count=$(( _count + 1 ))
	    fi
	}

	_old_right="$_right"
	_left="${_old_right%%"$_separator"*}"
	_right="${_old_right#*"$_separator"}"

	if [[ "$_left" == "$_old_right" ]]
	then # no separator found
	    bb_string_split__result+=("$_right")
	    _right=""
	else # separator found
	    bb_string_split__result+=("$_left")
	    [ -n "$_right" ] || bb_string_split__result+=("")
	fi
    done

    bb_array_copy "$_ret_array_vn" bb_string_split__result
}

# _bb_string_split_regex <ret-array-vn> <string> <regex> <limit> <store-mode>
function _bb_string_split_regex() {
    ((  $# == 5 )) || {
	bb_ierr "called with $# instead of 5 args"
	return 2
    }

    local _ret_array_vn="$1"
    local _string="$2"
    local _regex="$3"
    local _limit="${4-}"
    local _store_mode="$5"

    bb_assert_array_is_defined "$_ret_array_vn"

    [ -z "$_limit" ] || {
	bb_uint_validate "$_limit" || return 4
	(( _limit > 0 )) || return 4
    }

    local -a bb_string_split_regex__result=()
    local _left _right _m_m _m_r _m_l
    local _count=1

    _right="$_string"
    while (( ${#_right} > 0 ))
    do
	[ -z "$_limit" ] || {
	    if (( _count >= _limit ))
	    then
		# limit was hit
		bb_string_split_regex__result+=("$_right")
		break
	    else
		_count=$(( _count + 1 ))
	    fi
	}

	if [[ "$_right" =~ (${_regex})(.*$) ]]
	then
	    _m_m="${BASH_REMATCH[1]}"
	    _m_r="${BASH_REMATCH[2]}"
	    _m_l="${_right:0:$(( ${#_right} - ${#_m_m} - ${#_m_r} ))}"
	    case "$_store_mode" in
		subs) bb_string_split_regex__result+=("$_m_l");;
		full) bb_string_split_regex__result+=("$_m_l" "$_m_m");;
		*)
		    bb_ierr "illegal store-mode"
		    return 4
	    esac

	    # work, if regex is at end of string
	    [ -n "$_m_r" ] || bb_string_split_regex__result+=("$_m_r")

	    if [ "$_right" == "$_m_r" ]
	    then
		# we have to abort, so we do not loop forever
		return 91
	    fi
	    _right="$_m_r"
	else
	    # regex not found
	    bb_string_split_regex__result+=("$_right")
	    _right=""
	fi
    done

    bb_array_copy "$_ret_array_vn" bb_string_split_regex__result
}

#@ bb_string_split_regex <ret-array-vn> <string> <regex> [<limit>]
# splits string by regex
# the regex must not contain marked subexpressions and must not match the empty
# string
# if limit is specified then it splits into at most limit parts starting
# from the left side, limit must be a number > 0
# errors: 2, 3, 4, 91 (regex matches empty string)
function bb_string_split_regex() {
    ((  $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    local _ret_array_vn="$1"
    local _string="$2"
    local _regex="$3"
    local _limit="${4-}"

    _bb_string_split_regex "$_ret_array_vn" "$_string" "$_regex" "$_limit" subs
}

#@ bb_string_split_regex_full <ret-array-vn> <string> <regex> [<limit>]
# splits string by regex, but return delimiters as well as substrings
# the regex must not contain marked subexpressions and must not match the empty
# string
# if limit is specified then it splits into at most limit parts starting
# from the left side, limit must be a number > 0
# note that delimiters to not count as parts
# errors: 2, 3, 4, 91 (regex matches empty string)
function bb_string_split_regex_full() {
    ((  $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    local _ret_array_vn="$1"
    local _string="$2"
    local _regex="$3"
    local _limit="${4-}"

    _bb_string_split_regex "$_ret_array_vn" "$_string" "$_regex" "$_limit" full
}

# helper fun that creates lookup table for character transliterations
function _bb_char_transliterate_create_aa() {
    local _ucode='ÀÁÂÃÄÅàáâãäåĀāĂăĄąÇçĆćĈĉĊċČčÐðĎďĐđÈÉÊËèéêëĒēĔĕĖėĘęĚěĜĝĞğĠġĢģĤĥĦħÌÍÎÏìíîïĨĩĪīĬĭĮįİıĴĵĶķĸĹĺĻļĽľĿŀŁłÑñŃńŅņŇňŉŊŋÒÓÔÕÖØòóôõöøŌōŎŏŐőŔŕŖŗŘřŚśŜŝŞşŠšſŢţŤťŦŧÙÚÛÜùúûüŨũŪūŬŭŮůŰűŲųŴŵÝýÿŶŷŸŹźŻżŽž'
    local _ascii="AAAAAAaaaaaaAaAaAaCcCcCcCcCcDdDdDdEEEEeeeeEeEeEeEeEeGgGgGgGgHhHhIIIIiiiiIiIiIiIiIiJjKkkLlLlLlLlLlNnNnNnNnnNnOOOOOOooooooOoOoOoRrRrRrSsSsSsSssTtTtTtUUUUuuuuUuUuUuUuUuUuWwYyyYyYZzZzZz"

    declare -g -A _bb_char_transliterate__lookup_table
    local _i
    for ((_i=0; _i < ${#_ucode}; _i++ ))
    do
	_bb_char_transliterate__lookup_table[${_ucode:$_i:1}]="${_ascii:$_i:1}"
    done
}
_bb_char_transliterate_create_aa

#@ bb_char_transliterate <ret-vn> <char> [ascii|username]
# transliterates single char to a ASCII or usernames (ASCII is default)
# when username is used, chars are transliterated to
# ASCII Alphanumeric Chards and the following: _ . $ -
# note: this function returns strings of length 1 or 2
# errors: 2, 3, 4
function bb_char_transliterate() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "called with $# instead of 2 or 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _char="$2"
    local _charset="${3-ascii}"

    bb_assert_string_is_defined "$_ret_vn"

    (( ${#_char} == 1 )) || {
	bb_msg err "${FUNCNAME[0]-?} expected one char but got: $_char"
	return 4
    }

    local _charset_range
    # shellcheck disable=SC2034
    case "$_charset" in
	ascii) _charset_range='[:ascii:]';;
	username) _charset_range='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0-9_.$-';; # :alnum: does not here since it includes umlauts
	*)
	    bb_msg err "${FUNCNAME[0]-?} called with unknown charset: $_charset"
	    return 4
    esac

    local _ret_string
    # shellcheck disable=SC2102
    case "$_char" in
	[${_charset_range}]) _ret_string="$_char";;
	"ä") _ret_string="ae";;
	"ö") _ret_string="oe";;
	"ü") _ret_string="ue";;
	"Ä") _ret_string="AE";;
	"Ö") _ret_string="OE";;
	"Ü") _ret_string="UE";;
	"ß") _ret_string="ss";;
	"ẞ") _ret_string="SS";;
	*) _ret_string="${_bb_char_transliterate__lookup_table["$_char"]-_}"
    esac

    printf -v "$_ret_vn" "%s" "$_ret_string"
}

#@ bb_string_transliterate <ret-vn> <string> [ascii|username]
# transliterates each character of a string with char_transliterate
# optional argument is target charset, ascii is default
# see bb_char_transliterate for more on charsets
# note: the result might be longer than the input string but never cropped
# errors: 1, 2, 3, 4
function bb_string_transliterate() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "called with $# instead of 2 or 3 args"
	return 2
    }

    # shellcheck disable=SC2034
    local _ret_vn="$1"
    local _string="$2"
    local _charset="${3-ascii}"

    bb_assert_string_is_defined "$_ret_vn"

    [[ "$_charset" =~ ^((ascii)|(username))$ ]] || {
	bb_msg err "${FUNCNAME[0]-?} called with unknown charset: $_charset"
	return 4
    }

    # return "" if input string is empty
    [ -n "$_string" ] || {
	printf -v "$_ret_vn" "%s" ""
	return 0
    }

    # shellcheck disable=SC2034
    local -a bb_string_transliterate__array1 bb_string_transliterate__array2

    # break string into array, map it, then join again
    bb_array_of_string bb_string_transliterate__array1 "$_string" || return 1
    bb_array_map bb_string_transliterate__array2 \
	bb_string_transliterate__array1 bb_char_transliterate "$_charset" || \
	return 1
    bb_array_join_by "$_ret_vn" bb_string_transliterate__array2 "" || \
	return 1
}

#@@ regex related functions

#@ bb_regex_validate <regex>
# checks if string matches regex
# errors: 2, 3, 91 (invalid regex)
function bb_regex_validate() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    local _retval

    while (( $# > 0 ))
    do
	if [[ "" =~ $1 ]]
	then
	    continue # valid regex pattern (and matches)
	else
	    # this seems like a ShellCheck false positive
	    # shellcheck disable=SC2319
	    _retval="$?"
	    case "$_retval" in
		2) return 91;; # invalid regex pattern
		1) ;; # valid regex pattern (does not match)
		*)
		    bb_ierr "=~ returned unexpected value: $_retval"
		    return 1
	    esac
	fi
	shift
    done
}

#@ bb_regex_match <regex> <string>
# checks if string matches regex
# errors: 2, 3
function bb_regex_match() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    [[ "$2" =~ $1 ]] || return 3
}

#@@ array related functions

#@ bb_array_is_defined <vn>
# check if variable is defined and of type array
# errors: 2, 7, 11
function bb_array_is_defined() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local bb_array_is_defined__type
    bb_var_type bb_array_is_defined__type "" "" "$1" || return 7
    [ "$bb_array_is_defined__type" == "array" ] || return 11
}

#@ bb_array_is_set <vn>
# check if variable set (has been assigned a value) and of type array
# errors: 2, 7, 11, 12
function bb_array_is_set() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local bb_array_is_set__type bb_array_is_set__set
    bb_var_type bb_array_is_set__type "" bb_array_is_set__set "$1" || \
	return 7
    [ "$bb_array_is_set__type" == "array" ] || return 11
    [ "$bb_array_is_set__set" == "1" ] || return 12
}

#@ bb_array_is_empty <array-vn>
# check if array is set and empty
# errors:
# 2, 7 (array is undefined), 11 (defined but not an array),
# 12 (array is unset), 91 (not empty)
function bb_array_is_empty() {
    (( $# == 1 )) || {
	bb_ierr "${FUNCTION[0]-?} called with $# instead of 1 arg"
	return 2
    }

    bb_array_is_set "$1" || return "$?" # 7, 11, 12

    local -n _array_ref="$1"

    (( ${#_array_ref[@]} == 0 )) || return 91
}

#@ bb_array_contains_value <array-vn> <string>
# check if array contains value <string>, return 0 if found
# errors: 2, 3, 10
function bb_array_contains_value() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _array_vn="$1"
    local _str="$2"
    local -a bb_array_contains_value__fr

    bb_assert_array_is_set "$_array_vn"

    bb_array_filter bb_array_contains_value__fr "$_array_vn" \
	bb_string_compare "$_str"

    (( ${#bb_array_contains_value__fr[@]} > 0 )) || return 10
}

#@ bb_array_contains_key <array-vn> <key>
# check if array contains key
# errors: 2, 3, 10
function bb_array_contains_key() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _array_vn="$1"
    local _key="$2"

    bb_assert_array_is_set "$_array_vn"

    bb_uint_validate "$_key" || return 3

    local _ref
    _ref="$_array_vn""[$_key]"
    [ "${!_ref+defined}" == "defined" ] || return 10
}

#@ bb_array_compare <array1-vn> <array2-vn>
# deep compare two arrays, return 0 if identically
# errors: 2, 3, 91 (not identically)
function bb_array_compare() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _src1_vn="$1"
    local _src2_vn="$2"

    bb_assert_array_is_set "$_src1_vn"
    bb_assert_array_is_set "$_src2_vn"

    local _src1 _src2
    _src1="$(declare -p "$_src1_vn")"
    _src1="${_src1#*=}"
    _src2="$(declare -p "$_src2_vn")"
    _src2="${_src2#*=}"

    [[ "$_src1" == "$_src2" ]] || return 91
}

#@ bb_array_add <array-vn> <element> [..]
# add elements to set array
# errors: 2, 3
function bb_array_add() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    bb_assert_array_is_set "$1"

    local -n _dst_vn_ref="$1"
    shift

    _dst_vn_ref+=("$@")
}

#@ bb_array_concat <array-vn> <addon-array-vn>
# add contents of addon-array to exisg array-vn
# indices of addon-array will be ignored, next free indices of array are used
# errors: 2, 3
function bb_array_concat() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _array_vn="$1"
    local _addon_array_vn="$2"

    bb_assert_array_is_set "$_array_vn"
    bb_assert_array_is_set "$_addon_array_vn"

    if bb_array_is_empty "$_addon_array_vn"
    then # nothing to do if array is empty
	return 0
    fi

    local -n _array_ref="$_array_vn"
    local -n _addon_array_ref="$_addon_array_vn"

    _array_ref+=("${_addon_array_ref[@]}")
}

#@ bb_array_set <array-vn> [<element> [..]]
# overwrite contents of array
# for indirect version see bb_array_copy
# errors: 2, 3
function bb_array_set() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"

    local -n _dst_array_ref="$1"
    shift

    _dst_array_ref=("$@")
}

#@ bb_array_insert <array-vn> <key> <element>
# insert an element at specified key
# elements in way will recursivly be copied to key+1
# note: in sparse arrays not all keys > key might be touched
# errors: 2, 3
function bb_array_insert() {
    (( $# == 3 )) || {
        bb_ierr "called with $# instead of 3 args"
        return 2
    }

    local _array_vn="$1"
    local _key="$2"
    local _value="$3"

    bb_assert_array_is_defined "$_array_vn"

    local _ref _backup
    local -n _array_ref="$_array_vn"

    while true; do
        if [[ -n "${_array_ref[$_key]+exists}" ]]; then
            _backup="${_array_ref[$_key]}"
            _array_ref["$_key"]="$_value"
            _value="$_backup"
            _key=$((_key + 1))
        else
            _array_ref["$_key"]="$_value"
            return 0
        fi
    done
}

#@ bb_array_get <ret-vn> <array-vn> <key>
# returns element of array
# errors: 2, 3, 10
function bb_array_get() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _array_vn="$2"
    local _key="$3"

    bb_assert_string_is_defined "$_ret_vn"
    bb_assert_array_is_defined "$_array_vn"

    bb_uint_validate "$_key" || return 3

    local _ref
    _ref="$_array_vn""[$_key]"
    [ "${!_ref+defined}" == "defined" ] || return 10
    printf -v "$_ret_vn" "%s" "${!_ref}"
}

#@ bb_array_clear <array_vn>
# clear a already defined array
# errors: 2, 3
function bb_array_clear() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _dst_vn="$1"

    bb_assert_array_is_defined "$_dst_vn"

    bb_array_set "$_dst_vn"
}

#@ bb_array_copy_without_vals <dst-vn> <src-vn> <val> [..]
# copies array named srv-vn into array called dst-vn while skipping all
# specified vals
# errors: 2
function bb_array_copy_without_vals() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"
    shift 2

    bb_assert_array_is_defined "$_dst_vn"
    bb_assert_array_is_set "$_src_vn"

    local -n _src_ref="$_src_vn"
    local -n _dst_ref="$_dst_vn"

    local -a _src_keys
    _src_keys=("${!_src_ref[@]}")

    # clear destination array
    _dst_ref=()

    # empty without anything is empty
    (( ${#_src_keys[@]} > 0 )) || return 0

    local _key _val _element
    for _key in "${_src_keys[@]}"; do
        _val="${_src_ref[$_key]}"
        for _element in "$@"; do
            [[ "$_val" != "$_element" ]] || continue 2
        done
        _dst_ref["$_key"]="$_val"
    done
}

#@ bb_array_remove_vals <array-vn> <val> [..]
# removes specified vals from array named array-vn
# errors: 2, 3
function bb_array_remove_vals() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    local _array_vn="$1"
    shift

    bb_assert_array_is_set "$_array_vn"

    # shellcheck disable=SC2034
    local -a bb_array_remove_vals__tmp_a
    bb_array_copy_without_vals bb_array_remove_vals__tmp_a "$_array_vn" "$@"
    bb_array_copy "$_array_vn" bb_array_remove_vals__tmp_a
}

#@ bb_array_unsparse <ret-array-vn> <array-vn>
# returns non sparse version of input array, filled with by empty values
# filling will be done from index 0 to max index in source array
# e.g. ([1]=1 [3]=3 [6]=6) would be converted to ("" 1 "" 3 "" "" 6)
# errors: 2, 3, 7, 11
function bb_array_unsparse() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_assert_array_is_defined "$_dst_vn"
    bb_assert_array_is_set "$_src_vn"

    local -n _src_array_ref="$_src_vn"
    local -n _dst_array_ref="$_dst_vn"

    # clear destination array
    _dst_array_ref=()

    local -a _keys
    _keys=("${!_src_array_ref[@]}")

    # no need to process empty arrays
    (( ${#_keys[@]} > 0 )) || return 0

    local -i _i=0 _key
    for _key in "${_keys[@]}"
    do
	# fill sparse area with empty values
	for (( ; _i < _key; _i++ ))
	do
	    _dst_array_ref[_i]=""
	done

	# copy element
	_dst_array_ref[_i]="${_src_array_ref[_key]}"
	_i=$(( _i + 1 ))
    done
}

#@ bb_array_reindex <ret-array-vn> <array-vn>
# return version of array where keys are re-indexed starting with 0
# e.g. ([1]=1 [10]=2) will transform to ([0]=1 [1]=2)
# errors: 2
function bb_array_reindex() {
    (( $# == 2 )) || {
        bb_ierr "called with $# instead of 2 args"
        return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_assert_array_is_defined "$_dst_vn"
    bb_assert_array_is_set "$_src_vn"

    # Use namerefs for simplicity
    local -n _src_array_ref="$_src_vn"
    local -n _dst_array_ref="$_dst_vn"

    # Reindex the array
    _dst_array_ref=("${_src_array_ref[@]}")
}

#@ bb_array_compact <ret-array-vn> <array-vn>
# returns compacted array, all sparse areas will be removed
# e.g. ([0]=0 [2]=2) would be converted to ([0]=0 [1]=2)
# erors: 2, 3, 7, 11
function bb_array_compact() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_assert_array_is_defined "$_dst_vn"
    bb_assert_array_is_set "$_src_vn"

    local -n _src_array_ref="$_src_vn"
    local -n _dst_array_ref="$_dst_vn"

    local -a _keys
    _keys=("${!_src_array_ref[@]}")

    # clear destination array
    _dst_array_ref=()

    local -i _i _len="${#_keys[@]}" _key
    _len="${#_keys[@]}"
    for (( _i = 0; _i < _len; _i++ ))
    do
	_key="${_keys[$_i]}"

	_dst_array_ref[_i]="${_src_array_ref[_key]}"
    done
}

#@ bb_array_is_sparse <array-vn>
# checks if array is sparse
# errors: 1, 2, 3, 7, 11, 91 (array is not sparse)
function bb_array_is_sparse() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _array_vn="$1"

    bb_assert_array_is_set "$_array_vn"

    local -i _retval
    # shellcheck disable=SC2034
    local -a bb_array_is_sparse__a
    bb_array_compact bb_array_is_sparse__a "$_array_vn" || {
	_retval="$?"
	bb_ierr "failed to compact: $?"
	return 1
    }

    if bb_array_compare bb_array_is_sparse__a "$_array_vn"
    then # compact version matches, this array is not sparse
	return 91
    else
	_retval="$?"
	if (( _retval == 91 ))
	then # array is sparse
	    return 0
 	fi
	bb_ierr "failed to compare: $_retval"
	return 1
    fi
}

#@ bb_array_reverse <ret-array-vn> <array-vn>
# returns reversed array
# same indices are used, only data is swapped (0 with n, 1 with (n-1), ..)
# erors: 2, 3
function bb_array_reverse() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_assert_array_is_defined "$_dst_vn"
    bb_assert_array_is_set "$_src_vn"

    local -n _src_array_ref="$_src_vn"
    local -n _dst_array_ref="$_dst_vn"

    local -a _keys
    _keys=("${!_src_array_ref[@]}")

    # clear destination array
    _dst_array_ref=()

    local -i _i _j _ks _kt
    for (( _j = 0, _i = ${#_keys[@]} - 1; _i >= 0; _i--, _j++ ))
    do
	_ks="${_keys[$_i]}"
	_kt="${_keys[$_j]}"

	_dst_array_ref[_kt]="${_src_array_ref[_ks]}"
    done
}

# _bb_array_iter_x (d|k|i) <array-vn> <iter-exec> [<xarg> [..]]
function _bb_array_iter_x() {
    local _mode="$1"
    local _src_vn="$2"
    local _exec="$3"
    shift 3

    local -n _src_array_ref="$_src_vn"

    local -a _keys
    local _key _ref

    _keys=("${!_src_array_ref[@]}")

    # no need to process empty arrays
    (( ${#_keys[@]} > 0 )) || return 0

    local -i _retval=0 _index=0

    for _key in "${_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"
	case "$_mode" in
	    d) "$_exec" "${!_ref}" "$@" || _retval=91;;
	    k) "$_exec" "${!_ref}" "$_key" "$@" || _retval=91;;
	    i) "$_exec" "${!_ref}" "$_index" "$@" || _retval=91;;
	    *) return 1
	esac
	_index=$(( _index + 1 ))
    done

    return "$_retval"
}

#@ bb_array_iter <array-vn> <iter-exec> [<xarg> [..]]
# iter-exec must take: <element> [<xarg> [..]]
# note: iter-exec can be a fun or any other callable
# errors: 2, 3, 8, 91 (one or more exec calls failed)
function bb_array_iter() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    bb_assert_array_is_set "$1"
    bb_exec_is_callable "$2"

    _bb_array_iter_x d "$@" || return "$?"
}

#@ bb_array_iter_i <array-vn> <iter-exec> [<xarg> [..]]
# iter-exec must take: <element> <index> [<xarg> [..]]
# note: iter-exec can be a fun or any other callable
# note: this function processes elements one by one and ignores keys
# errors: 2, 3, 8, 91 (one or more exec calls failed)
function bb_array_iter_i() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    bb_assert_array_is_set "$1"
    bb_exec_is_callable "$2"

    _bb_array_iter_x i "$@" || return "$?"
}

#@ bb_array_iter_k <array-vn> <iter-exec> [<xarg> [..]]
# iter-exec must take: <element> <key> [<xarg> [..]]
# note: iter-exec can be a fun or any other callable
# errors: 2, 3, 8, 91 (one or more exec calls failed)
function bb_array_iter_k() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    bb_assert_array_is_set "$1"
    bb_exec_is_callable "$2"

    _bb_array_iter_x k "$@" || return "$?"
}

# _bb_array_iter2_x (d|k|i) <array1-vn> <array2-vn> <iter-exec> [<xarg> [..]]
function _bb_array_iter2_x() {
    local _mode="$1"
    local _src1_vn="$2"
    local _src2_vn="$3"
    local _exec="$4"
    shift 4

    local -n _src1_array_ref="$_src1_vn"
    local -n _src2_array_ref="$_src2_vn"

    local -a _keys1
    local -a _keys2

    _keys1=("${!_src1_array_ref[@]}")
    _keys2=("${!_src2_array_ref[@]}")

    (( ${#_keys1[@]} == ${#_keys2[@]} )) || return 91

    # no need to process empty arrays
    (( ${#_keys1[@]} > 0 )) || return 0

    local -i _retval=0
    local _index _ref1 _ref2 _key1 _key2
    # shellcheck disable=SC2034
    local bb_array_iter__cache
    for (( _index=0; _index < ${#_keys1[@]}; _index++ ))
    do
	_key1="${_keys1[$_index]}"
	_key2="${_keys2[$_index]}"
	_ref1="${_src1_vn}[$_key1]"
	_ref2="${_src2_vn}[$_key2]"

	case "$_mode" in
	    d) _mf_varargs=("${!_ref1}" "${!_ref2}" "$@");;
	    i) _mf_varargs=("${!_ref1}" "${!_ref2}" "$_index" "$@");;
	    k) _mf_varargs=("${!_ref1}" "${!_ref2}" "$_key1" "$_key2" "$@");;
	    *) return 1
	esac

	if ! "$_exec" "${_mf_varargs[@]}"
	then
	    _retval=92
	fi
    done

    return "$_retval"
}

#@ bb_array_iter2 <array1-vn> <array2-vn> <iter-exec> [<xarg> [..]]
# iter-exec must take: <element1> <element2> [<xarg> [..]]
# note: iter-exec can be a fun or any other callable
# errors: 2, 3, 8, 91 (one or more exec calls failed)
function bb_array_iter2() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_set "$1"
    bb_assert_array_is_set "$2"
    bb_exec_is_callable "$3"

    _bb_array_iter2_x d "$@" || return "$?"
}

#@ bb_array_iter2_i <array1-vn> <array2-vn> <iter-exec> [<xarg> [..]]
# iter-exec must take: <element1> <element2> <index> [<xarg> [..]]
# note: iter-exec can be a fun or any other callable
# note: this function processes elements one by one and ignores keys
# errors: 2, 3, 8, 91 (one or more exec calls failed)
function bb_array_iter2_i() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_set "$1"
    bb_assert_array_is_set "$2"
    bb_exec_is_callable "$3"

    _bb_array_iter2_x i "$@" || return "$?"
}

#@ bb_array_iter2_k <array1-vn> <array2-vn> <iter-exec> [<xarg> [..]]
# iter-exec must take: <element1> <element2> <key> [<xarg> [..]]
# note: iter-exec can be a fun or any other callable
# errors: 2, 3, 8, 91 (one or more exec calls failed)
function bb_array_iter2_k() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_set "$1"
    bb_assert_array_is_set "$2"
    bb_exec_is_callable "$3"

    _bb_array_iter2_x k "$@" || return "$?"
}


# _bb_array_map (d|k|i) <ret-array-vn> <array-vn> <map-fun> [<xarg> [..]]
function _bb_array_map_x() {
    local _mode="$1"
    local _dst_vn="$2"
    local _src_vn="$3"
    local _map_fun="$4"
    shift 4

    local -n _dst_array_ref="$_dst_vn"
    local -n _src_array_ref="$_src_vn"

    local -a _keys
    local bb_array_map__cache _key _ref

    _keys=("${!_src_array_ref[@]}")

    # clear destination array
    _dst_array_ref=()

    # no need to process empty arrays
    (( ${#_keys[@]} > 0 )) || return 0

    local -i _retval=0 _index=0
    local -a _mf_varargs
    for _key in "${_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"

	case "$_mode" in
	    d) _mf_varargs=("${!_ref}" "$@");;
	    i) _mf_varargs=("${!_ref}" "$_index" "$@");;
	    k) _mf_varargs=("${!_ref}" "$_key" "$@");;
	    *) return 1
	esac

	if "$_map_fun" bb_array_map__cache "${_mf_varargs[@]}"
	then
	    _dst_array_ref[_key]="${bb_array_map__cache}"
	else
	    _retval=91
	fi
	_index=$(( _index + 1 ))
    done

    return "$_retval"
}

#@ bb_array_map <ret-array-vn> <array-vn> <map-fun> [<xarg> [..]]
# map-fun must take: <ret-vn> <src-value> [<xarg> [..]]
# errors: 2, 3, 8, 91 (one or more mapfun calls failed)
function bb_array_map() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_map_x d "$@" || return "$?"
}

#@ bb_array_map_i <ret-array-vn> <array-vn> <map-fun> [<xarg> [..]]
# map-fun must take: <ret-vn> <src-value> <index> [<xarg> [..]]
# errors: 2, 3, 8, 91 (one or more mapfun calls failed)
function bb_array_map_i() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_map_x i "$@" || return "$?"
}

#@ bb_array_map_k <ret-array-vn> <array-vn> <map-fun> [<xarg> [..]]
# map-fun must take: <ret-vn> <src-value> <key> [<xarg> [..]]
# errors: 2, 3, 8, 91 (one or more mapfun calls failed)
function bb_array_map_k() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_map_x k "$@" || return "$?"
}

# _bb_array_map2_x <mode> <ret-array-vn> <array1-vn> <array2-vn>
#                  <map-fun> [<xarg> [..]]
function _bb_array_map2_x() {
    local _mode="$1"
    local _dst_vn="$2"
    local _src1_vn="$3"
    local _src2_vn="$4"
    local _map_fun="$5"
    shift 5

    local -n _dst_array_ref="$_dst_vn"
    local -n _src1_array_ref="$_src1_vn"
    local -n _src2_array_ref="$_src2_vn"

    local -a _keys1
    local -a _keys2
    _keys1=("${!_src1_array_ref[@]}")
    _keys2=("${!_src2_array_ref[@]}")

    (( ${#_keys1[@]} == ${#_keys2[@]} )) || return 91

    # clear destination array
    _dst_array_ref=()

    # no need to process empty arrays
    (( ${#_keys1[@]} > 0 )) || return 0

    local -i _retval=0
    local _index _ref1 _ref2 _key1 _key2
    # shellcheck disable=SC2034
    local bb_array_map__cache
    local -a _mf_varargs
    for (( _index=0; _index < ${#_keys1[@]}; _index++ ))
    do
	_key1="${_keys1[$_index]}"
	_key2="${_keys2[$_index]}"
	_ref1="${_src1_vn}[$_key1]"
	_ref2="${_src2_vn}[$_key2]"

	case "$_mode" in
	    d) _mf_varargs=("${!_ref1}" "${!_ref2}" "$@");;
	    i) _mf_varargs=("${!_ref1}" "${!_ref2}" "$_index" "$@");;
	    k) _mf_varargs=("${!_ref1}" "${!_ref2}" "$_key1" "$_key2" "$@");;
	    *) return 1
	esac

	if "$_map_fun" bb_array_map__cache "${_mf_varargs[@]}"
	then
	    _dst_array_ref[_index]="${bb_array_map__cache}"
	else
	    _retval=92
	fi
    done

    return "$_retval"
}

#@ bb_array_map2 <ret-array-vn> <array1-vn> <array2-vn> <map-fun>
#                [<xarg> [..]]
# map-fun must take: <ret-vn> <value1> <value2> <key1> <key2> [<xarg> [..]]
# note: this function processes elements one by one and ignores keys
# errors: 2, 3, 8, 91 (src1 and src2 have different length),
#         92 (one or more mapfun calls failed)
function bb_array_map2() {
    (( $# > 3 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_array_is_set "$3"
    bb_assert_fun_is_defined "$4"

    _bb_array_map2_x d "$@" || return $?
}

#@ bb_array_map2_i <ret-array-vn> <array1-vn> <array2-vn> <map-fun>
#                  [<xarg> [..]]
# map-fun must take: <ret-vn> <value1> <value2> <key1> <key2> [<xarg> [..]]
# note: this function processes elements one by one and ignores keys
# errors: 2, 3, 8, 91 (src1 and src2 have different length),
#         92 (one or more mapfun calls failed)
function bb_array_map2_i() {
    (( $# > 3 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_array_is_set "$3"
    bb_assert_fun_is_defined "$4"

    _bb_array_map2_x i "$@"
}

#@ bb_array_map2_k <ret-array-vn> <array1-vn> <array2-vn> <map-fun>
#                  [<xarg> [..]]
# map-fun must take: <ret-vn> <value1> <value2> <key1> <key2> [<xarg> [..]]
# errors: 2, 3, 8, 91 (src1 and src2 have different length),
#         92 (one or more mapfun calls failed)
function bb_array_map2_k() {
    (( $# > 3 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_array_is_set "$3"
    bb_assert_fun_is_defined "$4"

    _bb_array_map2_x k "$@"
}

# _bb_array_fold_right_x <mode> <ret-vn> <array-vn>
#                        <fold-fun> [<xarg> [..]]
function _bb_array_fold_right_x() {
    local _mode="$1"
    local _ret_vn="$2"
    local _src_vn="$3"
    local _fold_fun="$4"
    shift 4

    local -n _src_array_ref="$_src_vn"
    local -a bb_array_fold_right__keys

    # get keys from source array
    # shellcheck disable=SC2034
    bb_array_fold_right__keys=("${!_src_array_ref[@]}")

    # fold right wants to start from back
    local -a bb_array_fold_right__rkeys
    bb_array_reverse bb_array_fold_right__rkeys bb_array_fold_right__keys

    local bb_array_fold_right__cache _key _ref
    bb_array_fold_right__cache=""
    local -i _index=0
    local -a _mf_varargs
    for _key in "${bb_array_fold_right__rkeys[@]}"
    do
	_ref="${_src_vn}[$_key]"

	case "$_mode" in
	    d) _mf_varargs=("$bb_array_fold_right__cache" "$@");;
	    i) _mf_varargs=("$bb_array_fold_right__cache" "$_index" "$@");;
	    k) _mf_varargs=("$bb_array_fold_right__cache" "$_key" "$@");;
	    *) return 1
	esac

	"$_fold_fun" bb_array_fold_right__cache \
	    "${!_ref}" "${_mf_varargs[@]}" || return 91
        _index=$(( _index + 1 ))
    done

    printf -v "$_ret_vn" "%s" "$bb_array_fold_right__cache"
}

#@ bb_array_fold_right <ret-vn> <array-vn> <fold-fun> [<xarg> [..]]
# folds an array from right to left
# fold-fun must take: <ret-vn> <element> <prev-result> [<xarg> [..]]
# errors: 2, 3, 8, 91 (fold-fun call failed)
function bb_array_fold_right() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_string_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_fold_right_x d "$@" || return "$?"
}

#@ bb_array_fold_right_i <ret-vn> <array-vn> <fold-fun> [<xarg> [..]]
# folds an array from right to left
# fold-fun must take: <ret-vn> <element> <prev-result> <index> [<xarg> [..]]
# errors: 2, 3, 8, 91 (fold-fun call failed)
function bb_array_fold_right_i() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_string_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_fold_right_x i "$@" || return "$?"
}

#@ bb_array_fold_right_k <ret-vn> <array-vn> <fold-fun> [<xarg> [..]]
# folds an array from right to left
# fold-fun must take: <ret-vn> <element> <prev-result> <index> [<xarg> [..]]
# errors: 2, 3, 8, 91 (fold-fun call failed)
function bb_array_fold_right_k() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_string_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_fold_right_x k "$@" || return "$?"
}

# _bb_array_fold_left_x <mode> <ret-vn> <array-vn>
#                       <fold-fun> [<xarg> [..]]
function _bb_array_fold_left_x() {
    local _mode="$1"
    local _ret_vn="$2"
    local _src_vn="$3"
    local _fold_fun="$4"
    shift 4

    local -a _keys
    local -n _src_array_ref="$_src_vn"
    local bb_array_fold_left__cache _key
    local _ref

    # get keys from source array
    _keys=("${!_src_array_ref[@]}")

    bb_array_fold_left__cache=""
    local -i _index=0
    local -a _mf_varargs
    for _key in "${_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"

	case "$_mode" in
	    d) _mf_varargs=("$bb_array_fold_left__cache" "$@");;
	    i) _mf_varargs=("$bb_array_fold_left__cache" "$_index" "$@");;
	    k) _mf_varargs=("$bb_array_fold_left__cache" "$_key" "$@");;
	    *) return 1
	esac

	"$_fold_fun" bb_array_fold_left__cache \
	    "${!_ref}" "${_mf_varargs[@]}" || return 91
	_index=$(( _index + 1 ))
    done

    printf -v "$_ret_vn" "%s" "$bb_array_fold_left__cache"
}

#@ bb_array_fold_left <ret-vn> <array-vn> <fold-fun> [<arg> [..]]
# folds an array from left to right
# fold-fun must take: <ret-vn> <element> <prev-result> [<xarg> [..]]
# errors: 2, 3, 8, 91 (fold-fun call failed)
function bb_array_fold_left() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_string_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_fold_left_x d "$@" || return "$?"
}

#@ bb_array_fold_left_i <ret-vn> <array-vn> <fold-fun> [<arg> [..]]
# folds an array from left to right
# fold-fun must take: <ret-vn> <element> <prev-result> [<xarg> [..]]
# errors: 2, 3, 8, 91 (fold-fun call failed)
function bb_array_fold_left_i() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_string_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_fold_left_x i "$@" || return "$?"
}

#@ bb_array_fold_left_k <ret-vn> <array-vn> <fold-fun> [<arg> [..]]
# folds an array from left to right
# fold-fun must take: <ret-vn> <element> <prev-result> [<xarg> [..]]
# errors: 2, 3, 8, 91 (fold-fun call failed)
function bb_array_fold_left_k() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_string_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_fun_is_defined "$3"

    _bb_array_fold_left_x k "$@" || return "$?"
}

# _bb_array_filter (d|k|i) <array-vn> <filterfun-name> [<xarg> [..]]
function _bb_array_filter_x() {
    local _mode="$1"
    local _dst_vn="$2"
    local _src_vn="$3"
    local _filter_fun="$4"
    shift 4

    local -a _keys
    local _key _ref
    local -n _src_array_ref="$_src_vn"
    local -n _dst_array_ref="$_dst_vn"

    # clear destination array
    _dst_array_ref=()

    # get keys from source array
    _keys=("${!_src_array_ref[@]}")

    # no need to process empty arrays
    (( ${#_keys[@]} > 0 )) || return 0

    local -i _index=0
    local -a _mf_varargs
    for _key in "${_keys[@]}"
    do
        _ref="${_src_vn}[$_key]"

        case "$_mode" in
            d) _mf_varargs=("${!_ref}" "$@");;
            i) _mf_varargs=("${!_ref}" "$_index" "$@");;
            k) _mf_varargs=("${!_ref}" "$_key" "$@");;
            *) return 1
        esac

        if "$_filter_fun" "${_mf_varargs[@]}"
        then
	    _dst_array_ref[_key]="${!_ref}"
        fi
        _index=$(( _index + 1 ))
    done
}

#@ bb_array_filter <ret-array-vn> <array-vn> <filter-exec> [<xarg> [..]]
# filter-exec must take: <element> [<xarg> [..]]
# note: filter-exec can be a fun or any other callable
# and shall return 0 if the element is to be kept in output array
# errors: 2, 3, 8
function bb_array_filter() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_exec_is_callable "$3"

    _bb_array_filter_x d "$@" || return "$?"
}

#@ bb_array_filter_i <ret-array-vn> <array-vn> <filter-fun> [<xarg> [..]]
# filterfun must take: <element> <index> [<xarg> [..]]
# and shall return 0 if the element is to be kept in output array
# errors: 2, 3, 8
function bb_array_filter_i() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_exec_is_callable "$3"

    _bb_array_filter_x i "$@" || return "$?"
}

#@ bb_array_filter_k <ret-array-vn> <array-vn> <filter-fun> [<xarg> [..]]
# filterfun must take: <element> <key> [<xarg> [..]]
# and shall return 0 if the element is to be kept in output array
# errors: 2, 3, 8
function bb_array_filter_k() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    bb_assert_array_is_set "$2"
    bb_assert_exec_is_callable "$3"

    _bb_array_filter_x k "$@" || return "$?"
}

#@ bb_array_copy <ret-dst-array-vn> <src-array-vn>
# copys an array, target array must exist and if non empty will be merged
# errors: same as bb_array_map
function bb_array_copy() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_assert_array_is_defined "$_dst_vn"
    bb_assert_array_is_set "$_src_vn"


    bb_array_map "$_dst_vn" "$_src_vn" bb_mapfun_identity
}

# _bb_array_unique_helper <element>
function _bb_array_unique_helper() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _e="$1"

    if [ -n "${_tmp_aa["$_e"]+exists}" ]
    then
	# double element, skip
	return 1
    else
	# new element
	_tmp_aa["$_e"]=1
	return 0
    fi
}

#@ bb_array_unique <ret-dst-array-vn> <src-array-vn>
# copies array but skip non unique elements
# errors: 2, 3
function bb_array_unique() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_assert_array_is_defined "$_dst_vn"
    bb_assert_array_is_set "$_src_vn"

    # shellcheck disable=SC2034
    local -A _tmp_aa=()
    bb_array_filter "$_dst_vn" "$_src_vn" _bb_array_unique_helper
}

#@ bb_array_length <ret-vn> <array-vn>
# returns the length of an array
# errors: 2, 3
function bb_array_length() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _src_vn="$2"

    bb_assert_string_is_defined "$_ret_vn"
    bb_assert_array_is_set "$_src_vn"

    local -i _length
    if [[ -v "$_src_vn" ]]
    then
	local -n _src_ref="$_src_vn"
	_length="${#_src_ref[@]}"
    else
	_length=0
    fi

    printf -v "$_ret_vn" "%i" "$_length"
}

#@ bb_mapfun_identity <ret-vn> <value>
# the identity fun
# errors: 2, 3
function bb_mapfun_identity() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _value="$2"

    bb_assert_string_is_defined "$_ret_vn"

    printf -v "$_ret_vn" "%s" "$_value"
}

# _bb_array_join_by_fold_fun <ret-vn> <element> <prev-value> <index> <join-str>
# errors: 2
function  _bb_array_join_by_fold_fun() {
    (( $# == 5 )) || {
	bb_ierr "called with $# instead of 5 args"
	return 2
    }

    local _ret_vn="$1"
    local _element="$2"
    local _prev_result="$3"
    local _index="$4"
    local _join_str="$5"

    # no join string on first element
    (( _index > 0 )) || _join_str=""
    printf -v "$_ret_vn" "%s%s%s" "$_prev_result" "$_join_str" "$_element"
}

#@ bb_array_join_by <ret-vn> <array-vn> <join-string>
# joins array element with join-string, returns "" if array is empty
# errors: 1, 2, 3, 7, 11
function bb_array_join_by() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _array_vn="$2"
    local _join_str="$3"

    bb_assert_string_is_defined "$_ret_vn"
    bb_assert_array_is_set "$_array_vn"

    bb_array_fold_left_i "$_ret_vn" "$_array_vn" _bb_array_join_by_fold_fun \
	"$_join_str" || {
	bb_ierr "array_fold_left_i failed: $?"
	return 1
    }
}

#@ bb_array_join_by_direct <ret-vn> <join-string> [<element> [..]]
# joins array element with join-string, returns "" if no elements specified
# errors: 1, 2, 3, 7, 11
function bb_array_join_by_direct() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    local _ret_vn="$1"
    local _join_str="$2"
    shift 2

    bb_assert_string_is_defined "$_ret_vn"

    # return "" if no elements are specified
    (( $# > 0 )) || {
	printf -v "$_ret_vn" "%s" ""
	return 0
    }

    # shellcheck disable=SC2034
    local -a bb_array_join_by_direct__ar=("$@")
    bb_array_join_by "$_ret_vn" bb_array_join_by_direct__ar "$_join_str" || \
	return 1
}

#@ bb_array_of_string <ret-array-vn> <string> [ <batch-size> ]
# converts a string into an array of strings with size batch-size and copy
# it into the array called dst-vn
# the default batch-size is 1
# note: the last batch may be shorter than batch-size
# errors: 2. 3, 7, 11
function bb_array_of_string() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "called with $# instead of 2 or 3 args"
	return 2
    }

    local _dst_vn="$1"
    local _input_str="$2"
    local _batch_size="${3-1}"

    bb_assert_array_is_defined "$_dst_vn"

    local -n _dst_ref="$_dst_vn"

    # clear destination array
    _dst_ref=()

    # nothing to do when input string is empty
    (( ${#_input_str} > 0 )) || return 0

    local i c=0
    for (( i=0; i < ${#_input_str}; i = i + _batch_size, c++ )); do
        _dst_ref[c]="${_input_str:$i:$_batch_size}"
    done
}

#@ bb_array_sort_bubble <ret-array-vn> <array-vn> [string|int|fun] [<cmp-fun>]
# sort an array with bubblesort
# for more info see description of bb_array_sort
# errors: 2, 3, 4, 7, 11,
#         91 (element not int but int compare used), 92 (cmp-fun failed)
function bb_array_sort_bubble() {
    (( $# >= 2 && $# <= 4 )) || {
	bb_ierr "called with $# instead of 2, 3 or 4 args"
	return 2
    }

    local _dst_array_vn="$1"
    local _input_array_vn="$2"
    local _cmp_mode="${3-string}"

    case "$_cmp_mode" in
	string|int)
	    (( $# < 4 )) || {
		bb_ierr "cmp-fun specified but compare mode not fun"
		return 4
	    }
	    ;;
	fun)
	    (( $# == 4 )) || {
		bb_ierr "cmp-fun missing but compare mode is fun"
		return 4
	    }
	    local _cmp_fun="$4"
	    ;;
	*)
	    bb_ierr "called with illegal compare type: $_cmp_mode"
	    return 4
    esac

    bb_assert_array_is_defined "$_dst_array_vn"
    bb_assert_array_is_set "$_input_array_vn"

    local -a bb_array_sort_bubble__ca
    bb_array_compact bb_array_sort_bubble__ca "$_input_array_vn" || \
	return "$?"

    local -i bb_array_sort_bubble__len
    bb_array_length bb_array_sort_bubble__len bb_array_sort_bubble__ca || \
	reeturn "$?"

    local -i _n="$bb_array_sort_bubble__len" _i _swapped=1
    local _tmp

    # all variants are copied to keep inner loops fast
    case "$_cmp_mode" in
	string)
	    while ((_swapped == 1))
	    do
		_swapped=0
		for (( _i=0; _i < _n-1; _i++ ))
		do
		    if [[ "${bb_array_sort_bubble__ca[_i]}" \
			      > "${bb_array_sort_bubble__ca[_i + 1]}" ]]
		    then
			_tmp="${bb_array_sort_bubble__ca[_i]}"
			bb_array_sort_bubble__ca[_i]="${bb_array_sort_bubble__ca[_i + 1]}"
			bb_array_sort_bubble__ca[_i + 1]="$_tmp"
			_swapped=1
		    fi
		done
		_n=$(( _n - 1 ))
	    done
	    ;;
	int)
	    bb_int_validate "${bb_array_sort_bubble__ca[@]}" || return 91

	    while ((_swapped == 1))
	    do
		_swapped=0
		for (( _i=0; _i < _n-1; _i++ ))
		do
		    if (( bb_array_sort_bubble__ca[_i] >
			  bb_array_sort_bubble__ca[_i + 1] ))
		    then
			_tmp="${bb_array_sort_bubble__ca[_i]}"
			bb_array_sort_bubble__ca[_i]="${bb_array_sort_bubble__ca[_i + 1]}"
			bb_array_sort_bubble__ca[_i + 1]="$_tmp"
			_swapped=1
		    fi
		done
		_n=$(( _n - 1 ))
	    done
	    ;;
	fun)
	    local -i bb_array_sort_bubble__cmp_retcode

	    while ((_swapped == 1))
	    do
		_swapped=0
		for (( _i=0; _i < _n-1; _i++ ))
		do
		    "$_cmp_fun" bb_array_sort_bubble__cmp_retcode \
				"${bb_array_sort_bubble__ca[_i]}" \
				"${bb_array_sort_bubble__ca[_i + 1]}" || \
			return 92
		    if (( bb_array_sort_bubble__cmp_retcode > 0 ))
		    then
			_tmp="${bb_array_sort_bubble__ca[_i]}"
			bb_array_sort_bubble__ca[_i]="${bb_array_sort_bubble__ca[_i + 1]}"
			bb_array_sort_bubble__ca[_i + 1]="$_tmp"
			_swapped=1
		    fi
		done
		_n=$(( _n - 1 ))
	    done
	    ;;

    esac

    bb_array_copy "$_dst_array_vn" bb_array_sort_bubble__ca
}

#@ bb_array_sort <ret-array-vn> <array-vn> [string|int|fun] [<cmp-fun>]
# sort an array
# compare method can be selected, string is default
# cmp-fun must take: <ret-code-vn> <element1> <element2>
#  and return 0 on success, ret-code-vn must be 0 if elements compare as equal,
#  a positive integer if the first is greater, and a negative integer if the
#  first is smaller
# array will be compacted before sorting
# sort algorithm will be chosen by bash-base
# errors: 2, 3, 4, 7, 11,
#         91 (element not int but int compare used), 92 (cmp-fun failed)
function bb_array_sort() {
    bb_array_sort_bubble "$@"
}

#@@ dict (associative array) related functions

#@ bb_dict_is_defined <vn>
# check if variable is defined and of type dict
# errors: 2, 7, 11
function bb_dict_is_defined() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local bb_dict_is_defined__type
    bb_var_type bb_dict_is_defined__type "" "" "$1" || return 7
    [ "$bb_dict_is_defined__type" == "dict" ] || return 11
}

#@ bb_dict_is_set <vn>
# check if variable set (has been assigned a value) and of type dict
# errors: 2, 7, 11, 12
function bb_dict_is_set() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local bb_dict_is_set__type bb_dict_is_set__set
    bb_var_type bb_dict_is_set__type "" bb_dict_is_set__set "$1" || \
	return 7
    [ "$bb_dict_is_set__type" == "dict" ] || return 11
    [ "$bb_dict_is_set__set" == "1" ] || return 12
}

#@ bb_dict_is_empty <dict-vn>
# check if dict is empty
# 2, 7 (dict is undefined), 11 (defined but not a dict),
# 12 (dict is unset), 91 (not empty)
function bb_dict_is_empty() {
    (( $# == 1 )) || {
        bb_ierr "called with $# instead of 1 arg"
        return 2
    }

    bb_dict_is_set "$1" || return "$?" # 7, 11, 12

    local -n _dict_ref="$1"

    (( ${#_dict_ref[@]} == 0 )) || return 91
}

#@ bb_dict_contains_key <dict-vn> <key>
# checks if dict contains specified key
# errors: 2, 91 (key not found)
function bb_dict_contains_key() {
    (( $# == 2 )) || {
        bb_ierr "called with $# instead of 2 args"
        return 2
    }

    local _dict_vn="$1"
    local _key="$2"

    bb_assert_dict_is_set "$_dict_vn"

    local -n _dict_ref="$_dict_vn"

    if [[ -v _dict_ref["$_key"] ]]
    then
        return 0 # key exists
    else
        return 91 # key does not exist
    fi
}

#@ bb_dict_keys <ret-array-vn> <dict-vn>
# stores keys of dict to ret-keys-array
# errors: 2, 3
function bb_dict_keys() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    bb_assert_array_is_defined "$1"
    local -n _ret_keys_ref="$1"
    bb_assert_dict_is_set "$2"
    local -n _dict_ref="$2"

    _ret_keys_ref=("${!_dict_ref[@]}")
}

#@ bb_dict_get_value <ret-vn> <dict-vn> <key> [<default>]
# if key is found in a dict, return its value
# if key is not found return default if specified or retval 10 otherwise
# errors: 2, 91 (key not found and no default specified)
function bb_dict_get_value() {
    (( $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    bb_assert_string_is_defined "$1"
    local -n _ret_ref="$1"
    bb_assert_dict_is_set "$2"
    local -n _dict_ref="$2"
    local _key="$3"

    if [[ -v _dict_ref["$_key"] ]]
    then
	# key exists
	_ret_ref="${_dict_ref["$_key"]}"
    else
	# key does not exist
	(( $# == 4 )) || return 91 # no default defined
        _ret_ref="$4"
    fi
}

#@ bb_dict_compare <dict1-vn> <dict2-vn>
# compares two dicts
# note: order of keys is ignored
# errors: 2, 3, 91 (not identically)
function bb_dict_compare() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    bb_assert_dict_is_set "$1"
    bb_assert_dict_is_set "$2"

    local -n _dict1_ref="$1"
    local -n _dict2_ref="$2"

    # fail if size does not match
    (( ${#_dict1_ref[@]} == ${#_dict2_ref[@]} )) || return 91

    # iterate over all keys in dict1
    # if key is not present in dict2, fail
    # if key is present compare values, fail if they differ
    local _key
    for _key in "${!_dict1_ref[@]}"
    do
        [[ -v _dict2_ref["$_key"] ]] || return 91
        [[ "${_dict1_ref["$_key"]}" == "${_dict2_ref["$_key"]}" ]] || return 91
    done
    # no need to do the same for dict2 since we compared sizes
}

#@ bb_dict_set <dict-vn> <key> <value>
# set key value par to set dict
# note: if key exists value will be overwritten
# errors: 2, 3
function bb_dict_set() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    bb_assert_dict_is_set "$1"

    local -n _dict_vn_ref="$1"

    _dict_vn_ref["$2"]="$3"
}

#@@ hexadecimal related functions

#@ bb_hex_validate <hex-string> [..]
# validates one or more hex strings
# valid variable names match the following regex: ^[[:xdigit:]]*$
# errors: 2, 91 (illegal hex code)
function bb_hex_validate() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    while (( $# > 0 ))
    do
	[[ "$1" =~ ^[[:xdigit:]]*$ ]] || return 91
	shift
    done
}

#@ bb_hex_read <ret-vn>
# converts binary data read from stdin to its hex representation
# make sure it does not run in subshell, call it e.g.: bb_hex_read x < <(date)
# requires: od installed (e.g. /usr/bin/od)
# errors: 2, 3, 91 (od not found)
function bb_hex_read() {
    (( $# == 1 )) || {
	bb_ierr "called with $# args instead of 1"
	return 2
    }

    bb_exec_is_callable "od" || return 91

    local _ret_vn="$1"

    bb_assert_string_is_defined "$_ret_vn"

    bb_vn_validate "$_ret_vn" "bb_hex_read:ret-vn"|| return 3

    local _ret_str=
    while read -r line
    do
	printf -v _ret_str "%s%s" "$_ret_str" "${line//[[:space:]]/}"
    done < <(od -A n -v -t x1)

   printf -v "$_ret_vn" "%s" "$_ret_str"
}

#@ bb_bes_of_hex <ret-vn> <hex>
# converts hex to backslash escape sequence, which can be converted to data
# by passing to: printf "%b"
# hex is interpreted as stream of bytes (means it must be of even length)
# errors: 2, 91 (hex is of odd length)
function bb_hex_to_bes() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _input="$2"

    bb_assert_string_is_defined "$_ret_vn"

    local -n _ret_vn_ref="$_ret_vn"

    (( ${#_input} > 0 )) || { # no input leads to no output
	_ret_vn_ref=""
	return
    }

    (( ${#_input} % 2 == 0 )) || return 91

    # shellcheck disable=SC2034
    local -a bb_hex_to_bes__inar
    local bb_hex_to_bes__joined
    bb_array_of_string bb_hex_to_bes__inar "$_input" 2 || return "$?"
    bb_array_join_by bb_hex_to_bes__joined bb_hex_to_bes__inar "\\x" || \
	return "$?"

    _ret_vn_ref="\\x$bb_hex_to_bes__joined"
}

#@ bb_hex_write <hex>
# writes hex encoded data as binary data to stdout
# hex is interpreted as stream of bytes (means it must be of even length)
# errors: 2, 91 (hex is of odd length), 92 (invalid hex char)
function bb_hex_write() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _hex="$1"
    bb_hex_validate "$_hex" || return 92

    local bb_hex_write__bes

    bb_hex_to_bes bb_hex_write__bes "$_hex"
    printf "%b" "$bb_hex_write__bes"
}

#@ bb_hex_xor <ret-vn> <hex1> <hex2>
# returns hex1 xor hex2 as hex
# errors: 2, 3, 91 (input length vary), 92 (invaid hex char)
function bb_hex_xor() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _hex1="$2"
    local _hex2="$3"

    bb_assert_string_is_defined "$_ret_vn"

    bb_hex_validate "$_hex1" || return 92
    bb_hex_validate "$_hex2" || return 92

    local _len
    _len=${#_hex1}
    (( _len == ${#_hex2} )) || return 91

    local _i _ret_str=
    for (( _i=0; _i < _len; _i++ ))
    do
	printf -v _ret_str "%s%x" \
	    "$_ret_str" $(( 0x${_hex1:$_i:1} ^ 0x${_hex2:$_i:1} ))
    done

    printf -v "$_ret_vn" "%s" "$_ret_str"
}

#@@ lock files

#@ bb_lock_aquire <lock-name> <lock-file> <lock-type> <timeout>
#                [<callback> [<granularity>]]
# aquire a lock on lock-file (contents will be overwritten)
# lock-name can be used to release lock later on
# lock-type can be read or write
# if timeout is > 0 wait timeout seconds, 0 means for ever, -1 do not wait
# if a callback function is specified it will be called while waiting
# to get the lock with the current lock owners message as argument
# note: it can be called multiple times if the owner changes
# lock-file is checked every granularity seconds if the owner changed
# the default granularity is 10
# errors: 2, 4, 91 (failed to aquire lock), 92 (error calling flock),
#         93 (flock not found in path), 94 (failed to write to lockfile)
function bb_lock_aquire() {
    (( $# >= 4 && $# <= 6 )) || {
	bb_ierr "called with $# instead of 4 to 6 args"
	return 2
    }

    local _lock_name="$1"
    local _lock_file="$2"
    local _lock_mode="$3"
    local _timeout="$4"
    local _lock_callback="${5-}"
    local _sleep_granularity="${6-10}"

    bb_vn_validate "$_lock_name" "bb_lock_aquire:lock-name" || return 4
    bb_int_validate "$_timeout" || return 4
    bb_int_validate "$_sleep_granularity" || return 4

    local _flock_mode _lock_fd
    case "${_lock_mode}" in
	read)
	    # shellcheck disable=SC2093
	    exec {_lock_fd}< "$_lock_file"
	    _flock_mode=shared;;
	write)
	    # shellcheck disable=SC2093
	    exec {_lock_fd}>> "$_lock_file"
	    _flock_mode=exclusive;;
	*)
	    bb_msg err "${FUNCNAME[0]-?} called with unknown mode: $_lock_mode"
	    erturn 4
    esac


    flock --"$_flock_mode" --nonblock "$_lock_fd" || {
	# if flock returned an error, propagate it
	(( $? == 1 )) || {
	    exec {_lock_fd}>&- || \
		bb_msg err "${FUNCNAME[0]-?} failed to release lock on abort"
	    return 92
	}

	# we did not get the lock at first try, shall we wait?
	(( "$_timeout" != "-1" )) || {
	    exec {_lock_fd}>&- || \
		bb_msg err "${FUNCNAME[0]-?} failed to release lock on abort"
	    return 91 # waiting not allowed, abort
	}

	if [ -z "$_lock_callback" ]
	then
	    # simple solution without caring about callback
	    if (( _timeout == 0 ))
	    then
		# wait forever to get lock
		flock --"$_flock_mode" --wait "$_lock_fd" || {
		    exec {_lock_fd}>&- || bb_msg err "${FUNCNAME[0]-?}" \
			"failed to release lock on abort"
		    return 92 # must be error
		}
		# we got the lock!
	    else
		# wait for lock but not longer than specified timeout
		flock --"$_flock_mode" --timeout="$_timeout" "$_lock_fd" || {
		    exec {_lock_fd}>&- || bb_msg err "${FUNCNAME[0]-?}" \
			"failed to release lock on abort"
 		    (( $? == 1 )) || return 92
		    return 91
		}
		# we got the lock!
	    fi
	else
	    # callback is specified, check lock owner every granularity seconds
	    local _lock_owner _prev_lock_owner _t_start _flock_timeout
	    _prev_lock_owner=
	    _t_start="$SECONDS"

	    while true
	    do
		# check if we have to inform callback
		_lock_owner="$(head --lines=1 "$_lock_file")"
		[ "$_lock_owner" == "$_prev_lock_owner" ] || {
		    _prev_lock_owner="$_lock_owner"
		    "$_lock_callback" "$_lock_owner"
		}

		# check if timeout already occured and calculate next sleep
		if (( _timeout == 0 ))
		then
		    _flock_timeout="$_sleep_granularity"
		else
		    # calculate how long to sleep until we reach timeout
		    _flock_timeout=$((_t_start + _timeout - SECONDS))
		    (( _flock_timeout > 0 )) || {
			exec {_lock_fd}>&- || bb_msg err "${FUNCNAME[0]-?}" \
			    "failed to release lock on abort"
			return 91 # we are past timeout
		    }

		    # do not sleep longer than granularity
		    (( _flock_timeout <= _sleep_granularity )) || \
			_flock_timeout="$_sleep_granularity"
		fi

		# try to get lock
		if flock --"$_flock_mode" --timeout="$_flock_timeout" \
		    "$_lock_fd"
		then
		    # we got the lock
		    break
		else
		    (( $? == 1 )) || {
			exec {_lock_fd}>&- || bb_msg err "${FUNCNAME[0]-?}" \
			    "failed to release lock on abort"
			return 92 # error
		    }
		    # no lock yet, check for callback again
		fi
	    done
	fi
    }

    # we have the lock, do not close fd until the lock is released

    local _date_now
    _date_now="$(date +'%F %T %Z')" || _date_now="unknown time"

    # at this point we own the lock
    if [ "$_lock_mode" == "write" ]
    then
	printf "writelocked by %s:%s since %s\nnever delete this file\n" \
	    "$HOSTNAME" "$BASHPID" "$_date_now" > "$_lock_file" || {
	    bb_msg err "${FUNCNAME[0]-?} failed to write to lock-file:" \
		"$_lock_file"
	}
    fi

    bb__locks_mode["$_lock_name"]="$_lock_mode"
    bb__locks_file["$_lock_name"]="$_lock_file"
    bb__locks_file_fd["$_lock_name"]="$_lock_fd"
}
declare -A bb__locks_file bb__locks_file_fd bb__locks_mode

# bb_lock_relase <lock-name>
# unlocks a previous successfully locked lock-file
# errors: 2, 4, 91 (flock failed to unlock), 92 (no such active lock)
function bb_lock_release() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _lock_name="$1"

    bb_vn_validate "$_lock_name" || return 4

    [ "${bb__locks_mode["$_lock_name"]+x}" == "x" ] || return  92
    [ "${bb__locks_file["$_lock_name"]+x}" == "x" ] || return  92
    [ "${bb__locks_file_fd["$_lock_name"]+x}" == "x" ] || return  92
    local _lock_mode="${bb__locks_mode["$_lock_name"]}"
    local _lock_file="${bb__locks_file["$_lock_name"]}"
    local _lock_fd="${bb__locks_file_fd["$_lock_name"]}"

    local _date_now="unknown time"
    _date_now="$(date +'%F %T %Z')" || true

    if [ "$_lock_mode" == "write" ]
    then
	printf "free/read, last writer %s:%s until %s\n%s\n" \
	    "$HOSTNAME" "$BASHPID" "$_date_now" "never delete this file" \
	    > "$_lock_file"
    fi

    # release lock (just in case the following close fails)
    flock --unlock "$_lock_fd" || \
	bb_msg err "${FUNCNAME[0]-?} failed to unlock"

    # close lock-file fd
    exec {_lock_fd}>&- || \
	bb_msg err "${FUNCNAME[0]-?} failed to close lock-file"

    unset "bb__locks_mode[$_lock_name]"
    unset "bb__locks_file[$_lock_name]"
    unset "bb__locks_file_fd[$_lock_name]"
}

#@@ error handling: handle internal errors in a configurable way

#@ bb_ierr <msg> [..]
# call this on internal error, depending on mode it does:
#  mode "fatal": pass msg to bb_fatal
#  other mode: print error msg with specified loglevel
function bb_ierr() {
    local _msg

    if (( $# == 0 ))
    then
	_msg="${FUNCNAME[1]-?} internal error: called w/o message"
    else
	_msg="${FUNCNAME[1]-?} internal error: $*"
    fi

    if [[ "$bb__ierrmode" == fatal ]]
    then
	bb_fatal "$_msg"
    else
	bb_msg "$bb__ierrmode" "$_msg"
    fi
}

#@ bb_ierr_set_mode [fatal|<msg-loglevel>]
# default is "err"
function bb_ierr_set_mode() {
    local _m="${1-err}"

    if [[ "${_m,,}" =~ \
	^(fatal|debug[123]?|info|notice|warning|err|crit|alert|emerg)$ ]]
    then
	bb__ierrmode="${_m,,}"
    else
	bb_ierr "called with illegal argument: $_m"
    fi
}

#@ bb_fatal <message> [..]
# end program execution after logging message prepended with 'fatal error'
# since it produces some lines of output it is meant for internal errors
# only, if you want to abort your script with a message use bb_quit
function bb_fatal() {
    if (( $# == 0 ))
    then
	bb_msg emerg "unknown fatal error in ${FUNCNAME[1]-?}"
    else
	bb_msg emerg "fatal error in ${FUNCNAME[1]-?}: $*"
    fi
    bb_quit 1
}

#@@ msg: log/show messages

#@ bb_msg_levelname_of_string <ret-vn> <string>
# returns level as string if input was integer
# else returns string if it is a valid level
# errors: 2, 3, 4
function bb_msg_levelname_of_string() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _str="${2,,}"

    bb_assert_string_is_defined "$_ret_vn"

    if [[ "$_str" =~ ^[0-9]$ ]]
    then
	local -a _lookup=(emerg alert crit err warning notice info debug \
	    debug2 debug3)

	printf -v "$_ret_vn" "%s" "${_lookup[$_str]}"
    else
	if [[ "$_str" =~ \
	    ^(debug[123]?|info|notice|warning|err|crit|alert|emerg)$ ]]
	then
	    printf -v "$_ret_vn" "%s" "$_str"
	else
	    # invalid argument
	    return 4
	fi
    fi
}

#@ bb_msg_int_of_level <ret-vn> <level>
# convert level name to int or return int unmodified if in valid range
# errors: 2, 3, 4
function bb_msg_int_of_level() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _level="$2"

    bb_assert_string_is_defined "$_ret_vn"

    local _result

    if [[ "$_level" =~ ^[0-9]$ ]]
    then
	_result="$_level"
    else
	case "${_level,,}" in
	    emerg) _result=0;;
	    alert) _result=1;;
	    crit) _result=2;;
	    err) _result=3;;
	    warning) _result=4;;
	    notice) _result=5;;
	    info) _result=6;;
	    debug|debug1) _result=7;;
	    debug2) _result=8;;
	    debug3) _result=9;;
	    *) return 4
	esac
    fi

    printf -v "$_ret_vn" "%i" "$_result" # we are already int
}

#@ bb_msg_set_target <target-name> <type>[:<arg>] <min-level>[:<max-level]
#                   <prefix> [[no]trace]
# where type can be one of: stdout, stderr, file, syslog
# and arg is filename for file, facility for syslog
# level is one of: debug[123]?, info, notice, warning, err, crit, alert, emerg
# prefix can be: none, timestamp, prog, progpid, full
# if notrace is specified, no stack traces will be print to that target
# notrace is the default
# note: target-name must match regex ^[[:alpha:]][[:alnum:]_]*$
# errors: 2, 4
function bb_msg_set_target() {
    (( $# == 4 || $# == 5 )) || {
	bb_ierr "called with $# instead of 5 or 6 args"
	return 2
    }

    local _tn="$1"

    [[ "$_tn" =~ ^[[:alpha:]][[:alnum:]_]*$ ]] || {
	bb_msg err "${FUNCNAME[0]-?}: illegal target-name: $_tn"
	return 4
    }

    local _type="$2"
    local _levelstr="$3"
    local _prefix="$4"
    local _prtr="${5-notrace}"

    local bb_msg_set_target__minlevel bb_msg_set_target__maxlevel

    if [[ "$_levelstr" =~ : ]]
    then
	bb_msg_int_of_level bb_msg_set_target__minlevel "${_levelstr%%:*}" || {
	    bb_ierr "called with illegal min-level:" \
		"${_levelstr%%:*}"
	    return 4
	}
	bb_msg_int_of_level bb_msg_set_target__maxlevel "${_levelstr#*:}" || {
	    bb_ierr "called with illegal max-level:" \
		"${_levelstr#*:}"
	    return 4
	}
    else
	bb_msg_int_of_level bb_msg_set_target__minlevel "${_levelstr}" || {
	    bb_ierr "called with illegal min-level:" \
		"${_levelstr}"
	    return 4
	}
	bb_msg_set_target__maxlevel=0
    fi

    # extract arg from type if present
    local _arg=
    if [[ "$_type" =~ : ]]
    then
	_arg="${_type#*:}"
	_type="${_type%%:*}"
    fi

    case "$_type" in
	stdout|stderr)
	    [ -z "$_arg" ] || {
		bb_ierr "type must be empty for stdin/out"
		return 4
	    }
	    ;;
	file) ;;
	syslog)
	    if [ -z "$_arg" ]
	    then
		_arg=syslog # default to syslog
	    else
		[[ "$_arg" =~ ^(auth(priv)?|cron|daemon|ftp|lpr|mail|news|syslog|user|uucp|local[01234567]|)$ ]] || {
		    bb_ierr "illegal syslog arg: $_arg"
		    return 4
		}
	    fi
	    ;;
	*)
	    bb_ierr "called with illegal type: $_type"
	    return 4
    esac

    [[ "$_prefix" =~ ^(none|timestamp|prog|progpid|full)$ ]] || {
	bb_ierr "called with illegal prefix"
	return 4
    }

    case "$_prtr" in
	trace) _prtr=yes;;
	notrace) _prtr=no;;
	*)
	    bb_ierr "called with illegal arg 6: $_prtr"
	    return 4
    esac

    bb_msgtargets__type["$_tn"]="$_type"
    bb_msgtargets__arg["$_tn"]="$_arg"
    bb_msgtargets__min_level["$_tn"]="$bb_msg_set_target__minlevel"
    bb_msgtargets__max_level["$_tn"]="$bb_msg_set_target__maxlevel"
    bb_msgtargets__prefix["$_tn"]="$_prefix"
    bb_msgtargets__print_traces["$_tn"]="$_prtr"
}

#@ bb_msg_del_target <target-name>
# errors: 2
function bb_msg_del_target() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    unset "bb_msgtargets__type[$1]"
    unset "bb_msgtargets__arg[$1]"
    unset "bb_msgtargets__min_level[$1]"
    unset "bb_msgtargets__max_level[$1]"
    unset "bb_msgtargets__prefix[$1]"
    unset "bb_msgtargets__print_traces[$1]"
}

#@ bb_msg_clear_targets
# remove all message targets
# errors: 2
# shellcheck disable=SC2120
function bb_msg_clear_targets() {
    (( $# == 0 )) || {
	bb_ierr "called with $# instead of no arg"
	return 2
    }

    unset bb_msgtargets__type
    unset bb_msgtargets__arg
    unset bb_msgtargets__min_level
    unset bb_msgtargets__max_level
    unset bb_msgtargets__prefix
    unset bb_msgtargets__print_traces

    declare -g -A bb_msgtargets__type=()
    declare -g -A bb_msgtargets__arg=()
    declare -g -A bb_msgtargets__min_level=()
    declare -g -A bb_msgtargets__max_level=()
    declare -g -A bb_msgtargets__prefix=()
    declare -g -A bb_msgtargets__print_traces=()
}

# msgr helper functions take: <msg> <unused> <msg-level-string> <type-arg>
function _bb_msg_to_syslog() {
    logger -p "$4.$3l" "${2-}$1"
}
function _bb_msg_to_stdout() {
    printf "%s\n" "${2-}$1"
}
function _bb_msg_to_stderr() {
    printf "%s\n" "${2-}$1" >&2
}
function _bb_msg_to_file() {
    printf "%s\n" "${2-}$1" >> "$4"
}

#@ bb_msg <level> <string> [..]
# logs messages to configured targets
# note: will return 0, even if level is illegal
# errors: 2, 4
function bb_msg() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    local bb_msg__level bb_msg__levelname
    if bb_msg_int_of_level bb_msg__level "$1"
    then
	_level="$bb_msg__level"
    else
	bb_ierr "called with illegal message level: $1"
	return 4
    fi
    bb_msg_levelname_of_string bb_msg__levelname "$1" || {
	bb_ierr "called with illegal message level: $1"
	return 4
    }
    shift

    local _tname _targ _message_prefix="[UNKNOWN message prefix]"
    local _timestamp _prog _progpid
    _timestamp="$(date '+%F %T')"
    _prog="${0##*/}"
    _progpid="${_prog}[${BASHPID}]"

    for _tname in "${!bb_msgtargets__type[@]}"
    do
	(( "${bb_msgtargets__min_level["$_tname"]}" >= bb_msg__level )) || \
	    continue
	(( "${bb_msgtargets__max_level["$_tname"]}" <= bb_msg__level )) || \
	    continue

	case "${bb_msgtargets__prefix[$_tname]}" in
	    none) _message_prefix=;;
	    timestamp) _message_prefix="$_timestamp: ";;
	    prog) _message_prefix="$_prog: ";;
	    progpid) _message_prefix="$_progpid: ";;
	    full) _message_prefix="$_timestamp $_progpid: ";;
	esac

	_targ="${bb_msgtargets__arg[$_tname]}"

	local _type="${bb_msgtargets__type[$_tname]}"
	local _outfun
	if [[ "$_type" =~ ^syslog|std(out|err)|file$ ]]
	then
	    _outfun=_"bb_msg_to_$_type"
	    else
	    bb_ierr warning "encountered illegal type: $_type"
	fi
	"$_outfun" "$*" "$_message_prefix" "$bb_msg__levelname" "$_targ" ||  {
	    local _ref

	    _ref="${!0}"
	    printf "%s" "${_ref##*/}: bb_msg: failed to write message to file" \
		"($_targ): $_message_prefix$*" >&2
	}

	if [[ "${bb_msgtargets__print_traces[$_tname]}" == "yes" ]]
	then
	    # shellcheck disable=SC2034
	    local -a bb_msg__strace
	    bb_stacktrace_create bb_msg__strace 2
	    bb_array_iter bb_msg__strace "$_outfun" \
		"$_message_prefix" "$bb_msg__levelname" "$_targ"
	fi
    done
}

#@ bb_msg_iterator <msg> <level>
# iterator to output msgs in iterations
# errors: like bb_msg
function bb_msg_iterator() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _msg="$1"
    local _lvl="$2"

    bb_msg "$_lvl" "$_msg"
}

#@ bb_msg_stacktrace <level> [<skip-count>]
# outputs stack trace
# skips skip-count elements of stack, default is 2
# errors: 1, 2
function bb_msg_stacktrace() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _lvl="$1"
    local _skip_count="$2"

    # shellcheck disable=SC2034
    local -a bb_msg_stacktrace__strace
    bb_stacktrace_create bb_msg_stacktrace__strace "$_skip_count" || return 1
    bb_array_iter bb_msg_stacktrace__strace bb_msg_iterator "$_lvl" || \
	return 1
}

#@ bb_msg_from_stdin <level> <prefix>
# reads from fd 0 until EOF and calls bb_msg for each line read
# every line is prefixed with prefix
# will not do anything if fd 0 is connected to a terminal
# note: will stop if bb_msg fails and forward the return value
# errors: 2, 4
function bb_msg_from_stdin() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    if [ -t 0 ]
    then
	return
    fi

    local _line
    local _level="$1"
    local _prefix="$2"

    while read -r _line || [ -n "$_line" ]
    do
	bb_msg "$_level" "$_prefix$_line" || return "$?"
    done
}

#@ bb_msgf <level> <format> [<format-arg> [..]]
# logs printf style messages to configured targets
# note: will return 0, even if level is illegal
# errors: 2, 4
function bb_msgf() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    local _level="$1"
    local _format="$2"
    shift 2

    local _message
    # shellcheck disable=SC2059
    printf -v _message "$_format" "$@" || {
	bb_msg warning \
	    "${FUNCNAME[0]-?} illegal format-string($_format) or args: $*"
	return 4
    }
    bb_msg "$_level" "$_message"
}

#@ bb_stacktrace_create <ret-array-vn> [<skip>]
# create a human readable stacktrace, excluding this function and
# returns it as array of strings
# optional arg skip can be used to manually configure how many inner
# levels will not be displayed
# errors: 2, 3
function bb_stacktrace_create() {
    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _skip="${2-1}"

    bb_assert_array_is_defined "$_ret_vn"

    local -a bb_stacktrace_create__strace=()
    # shellcheck disable=SC2034
    local _i _str _intro="Execution at"

    for (( _i=_skip; _i < (${#FUNCNAME[@]} - 1); _i++ ))
    do
	printf -v _str "%s file \"%s\", line %i, function \"%s\"" "$_intro" \
	    "${BASH_SOURCE[$_i]}" "${BASH_LINENO[$_i-1]}" "${FUNCNAME[$_i]}"
	# shellcheck disable=SC2034
	bb_stacktrace_create__strace[_i-1]="$_str"
	_intro="Called from"
    done
    _i=$(( ${#FUNCNAME[@]} - 1 ))
    printf -v _str "%s file \"%s\", line %i" "Called from" \
	"${BASH_SOURCE[$_i]}" "${BASH_LINENO[$_i-1]}"
    # shellcheck disable=SC2034
    bb_stacktrace_create__strace[_i]="$_str"

    bb_array_copy "$_ret_vn" bb_stacktrace_create__strace
}

#@@ assert funs - wrappers for defined/set checkers with warnings on fail

#@ bb_assert_configure <mode> [<log-level>]
# configures what asserts do, depeding on mode:
#  0 .. don't check anything (like NDEBUG in C)
#  1 .. bb_msg error with log-level (warning by default) but continue
#       on failed assertion
#  2 .. bb_fatal on failed assertion
# note: the default mode is 1, on modes 0 and 2 log-level will be ignored
# errors: 2, 4
function bb_assert_configure() {
    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    if [[ "$1" =~ ^[012]$ ]]
    then
	if (( $1 == 1 ))
	then
	    local bb_assert_configure__msglvl

	    bb_msg_levelname_of_string bb_assert_configure__msglvl \
				       "${2-warning}" || {
		bb_ierr "called with illegal log-level: ${2-}"
		return 4
	    }

	    BASH_BASE_ASSERT_MSG_LEVEL="$bb_assert_configure__msglvl"
	else # mode 0 opr 2
	    unset BASH_BASE_ASSERT_MSG_LEVEL
	fi
	BASH_BASE_ASSERT_MODE="$1"
    else
	bb_ierr "called with illegal mode: $1"
	return 4
    fi
}
# default mode is 1 unless already someone already took care
[ -v BASH_BASE_ASSERT_MODE ] || BASH_BASE_ASSERT_MODE=1

# _bb_assert_warn [<extra-info>]
# helper fun that can be called to emit warning about failed assert
function _bb_assert_warn() {
    local _assert_fun="${FUNCNAME[1]-unknown}"
    local _caller="${FUNCNAME[2]-unknown}"
    local _caller_source_file="${BASH_SOURCE[2]-unknown}"
    local _caller_linenr="${BASH_LINENO[1]-unknown}"

    local _extra_info=""
    [ -z "${1-}" ] || _extra_info=": $1"
    local _msg
    printf -v _msg "assertion (%s) failed in %s at %s:%s%s" \
	   "$_assert_fun" "$_caller" "$_caller_source_file" "$_caller_linenr" \
	   "$_extra_info"

    if (( BASH_BASE_ASSERT_MODE > 1 ))
    then
	bb_fatal "$_msg"
    else
	bb_msg "${BASH_BASE_ASSERT_MSG_LEVEL-warning}" "$_msg"
    fi
}

#@ bb_assert_string_is_defined <vn> [<extra-info>]
# wrapper for bb_string_is_defined
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_string_is_defined
function bb_assert_string_is_defined() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_string_is_defined "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}

#@ bb_assert_string_is_set <vn> [<extra-info>]
# wrapper for bb_string_is_set
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_string_is_set
function bb_assert_string_is_set() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_string_is_set "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}

#@ bb_assert_array_is_defined <vn> [<extra-info>]
# wrapper for bb_array_is_defined
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_array_is_defined
function bb_assert_array_is_defined() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_array_is_defined "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}



#@ bb_assert_array_is_set <vn> [<extra-info>]
# wrapper for bb_array_is_set
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_array_is_set
function bb_assert_array_is_set() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_array_is_set "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}

#@ bb_assert_dict_is_defined <vn> [<extra-info>]
# wrapper for bb_dict_is_defined
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_dict_is_defined
function bb_assert_dict_is_defined() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_dict_is_defined "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}

#@ bb_assert_dict_is_set <vn> [<extra-info>]
# wrapper for bb_dict_is_set
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_dict_is_set
function bb_assert_dict_is_set() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_dict_is_set "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}

#@ bb_assert_exec_is_callable <exec-name> [<extra-info>]
# wrapper for bb_exec_is_callable
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_exec_is_callable
function bb_assert_exec_is_callable() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_exec_is_callable "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}

#@ bb_assert_fun_is_defined <fun-name> [<extra-info>]
# wrapper for bb_fun_is_defined
# outputs warning which includes location of assert and optional extra-info
# if wrapped function fails
# errors: see bb_fun_is_defined
function bb_assert_fun_is_defined() {
    (( ${BASH_BASE_ASSERT_MODE-1} > 0 )) || return 0 # asserts may be disabled

    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# instead of 1 or 2 args"
	return 2
    }

    local _retval

    bb_fun_is_defined "$1" || {
	_retval="$?"
	_bb_assert_warn "${2-}"
	return "$_retval"
    }
}

#@@ flow control and error handling

#@ bb_finalize
# unregisters all traps
function bb_finalize() {
    bb_msg debug3 "${FUNCNAME[0]-?} called"
    _bb_traps_disable
}

#@ bb_quit [<exit-code> [<log message>] [..]]
# the safe way to clean exit a script using bash_base
function bb_quit() {
    bb_finalize
    _bb_call_exit_hooks

    # no message
    if (( $# < 2 ))
    then
	bb_msg debug3 "${FUNCNAME[0]-?} called"
	exit "${1-0}"
    fi

    local _retval="$1"
    shift
    bb_msg info "quitting: $*"
    exit "$_retval"
}

#@ bb_exec <cmd> [arg [..]]
# execs cmd with args if specified
function bb_exec() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    bb_finalize
    _bb_call_exit_hooks

    bb_msg debug3 "${FUNCNAME[0]-?} $*"

    exec "$@" || bb_fatal "${FUNCNAME[0]-?} failed to exec: $*"
}

function _bb_err_exit_trap() {
    local _status=$?
    local _traptype="$1"

    case "$_traptype" in
	ERR|EXIT)
	    bb_msg emerg "${FUNCNAME[0]-?}: unexpected exit/error" \
		"near '$BASH_COMMAND', status=$_status, trap=$_traptype"
	    bb_quit "$_status"
	    ;;
	*)
	    bb_msg err "${FUNCNAME[0]-?}: internal error," \
		"unknown trap type: $_traptype"
    esac
}

# _bb_traps_setup
# sets up traps for ERR and EXIT to handle unexpected errors and exits
function _bb_traps_setup() {
    trap '_bb_err_exit_trap "ERR"' ERR
    trap '_bb_err_exit_trap "EXIT"' EXIT
}

# _bb_traps_disable
# disables traps
function _bb_traps_disable() {
    trap - ERR
    trap - EXIT
}

#@ bb_version <ret-vn>
# returns the version of bash base
# errors: 2, 3
function bb_version() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    bb_assert_string_is_defined "$1"

    printf -v "$1" "0.6.8"
}

#@ bb_version_check_loaded <version>
# return 0 if this base-base instance is >= version
# return 91 if not which means required version is not loaded
# errors: 2, 91
function bb_version_check_loaded() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    bb_version_compare "$1" "0.6.8"
    if (( $? == 2 ))
    then
	# required version is newer than we are :(
	return 91
    else
	return 0
    fi
}

#@ bb_version_compare <version1> <version2>
# compare version strings and return:
#  0 if they are equal
#  91 if version1 < version2
#  92 if version2 > version1
# note: currently supports version strings containing numbers and dots
#   this might change in a future version in a backward compatible way
# errors: 2, 4, 91, 92
function bb_version_compare() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 arg"
	return 2
    }

    [[ "$1" =~ ^[[:digit:].]+$ ]] || return 4
    [[ "$2" =~ ^[[:digit:].]+$ ]] || return 4

    local IFS='.'
    local -a _ver1 _ver2
    local _len _part1 _part2

    read -r -a _ver1 <<< "$1"
    read -r -a _ver2 <<< "$2"

    # Get the maximum _length of both versions to compare each _part
    _len=${#_ver1[@]}
    if [[ ${#_ver2[@]} -gt $_len ]]; then
	_len=${#_ver2[@]}
    fi

    # Compare each version _part
    for ((i=0; i<_len; i++)); do
	# If a _part is missing, treat it as 0
	_part1=${_ver1[i]:-0}
	_part2=${_ver2[i]:-0}

	# Compare numeric portions first, if possible
	if [[ "$_part1" =~ ^[0-9]+$ ]] && [[ "$_part2" =~ ^[0-9]+$ ]]
	then
	    if (( _part1 < _part2 )); then
		return 91  # Version 1 is older
	    elif (( _part1 > _part2 )); then
		return 92  # Version 1 is newer
	    fi
	fi
    done

    # versions are the equal
    return 0
}

# _bash_base_call_exit_hooks
# calls all exit hooks sorted by hooknames
function _bb_call_exit_hooks() {
    local _hookname

    # check if any hooks are registered
    (( ${#bb__exithooks[@]} > 0 )) || return 0

    for _hookname in "${!bb__exithooks[@]}"
    do
	bb_msg debug3 "${FUNCNAME[0]-?}: calling ${bb__exithooks[$_hookname]}"
	"${bb__exithooks[$_hookname]}" || \
	    bb_msg warning "exit hook '$_hookname' failed with exit code $?"
    done
}

#@ bb_register_exit_hook [ <hook-name> ] <function-name>
# registers a function to be called on exit, it is not possible to
# pass arguments to that function
# hooks will be called sorted by hooknames
# note: the default hook-name is the function-name
function bb_register_exit_hook() {
    (( $# == 1 || $# == 2 )) || {
	bb_ierr "called with $# insted of 1 or 2 args"
	return 2
    }

    local _hookname="$1"
    local _hookfun

    if (( $# == 2 ))
    then
	_hookfun="$2"
    else
	_hookfun="$_hookname"
    fi

    bb__exithooks["$_hookname"]="$_hookfun"
}

#@ bb_unregister_exit_hook <hookn-ame>
# removes a registered exit hook
# errors: 2
function bb_unregister_exit_hook() {
    (( $# == 1 )) || {
	bb_ierr "called with $# insted of 1 arg"
	return 2
    }

    local _hookname="$1"

    unset "bb__exithooks[$_hookname]"
}

#@ bb_load_lib <name>
# sources file for lib specified
# if name is no absolute, relative or local path it tries:
#   /usr/libexec/<name>.bash
#   /usr/libexec/<name>/<name>.bash
# absolute paths start with /
# local paths start with ./
# relative paths start with ../
# errors: 2, 5, 8, 91 (lib not found)
# shellcheck disable=SC1090
function bb_load_lib() {
    (( $# == 1 )) || {
	bb_ierr "called with $# insted of 1 arg"
	return 2
    }

    local _l="$1"

    if [[ "$_l" =~ ^\.?\.?/ ]] # is path absolute, local or relative?
    then # path is one of: absolute (starting with / or ./) or relative (../)
	if [ -r "$_l" ]
	then
	    source "$_l" || {
		bb_msg err "${FUNCNAME[0]-?}: error sourcing: $_l"
		return 5
	    }
	    return 0
	else
	    # no such file
	    return 8
	fi
    fi

    local _f
    for _f in "/usr/libexec/$_l.bash" "/usr/libexec/$_l/$_l.bash"
    do
	if [ -r "$_f" ]
	then
	    source "$_f" || {
		bb_msg err "error sourcing: $_f"
		return 5
	    }
	    return # library loaded successfully
	else
	    if [ -f "/usr/libexec/$_l.bash" ]
	    then
		bb_msg err "no access rights to: $_f"
		return 9
	    fi
	fi
    done

    bb_msg warning "${FUNCNAME[0]-?}: no such lib found: $_l"
    return 91
}

#@ bb_load_libs <name> [..]
# load one or more libs but continue if some fail to load
# errors: 2, 91 (not all libs loaded)
function bb_load_libs() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    local _retval=0

    while (( $# > 0 ))
    do
	bb_load_lib "$1" || _retval=91
	shift
    done

    return "$_retval"
}

#@ bb_require_libs <name> [..]
# load one or more libs and raise fatal error if any fail
# errors: 2, bb_fatal on any other error
function bb_require_libs() {
    (( $# > 0 )) || {
	bb_ierr "called with $# instead of 1 or more args"
	return 2
    }

    while (( $# > 0 ))
    do
	bb_load_lib "$1" || bb_fatal "${FUNCNAME[0]-?}: failed to load: $1"
	shift
    done
}

function _bb_main() {
    set -o errexit
    set -o pipefail
    set -o nounset
    set -o errtrace

    # default internal error mode
    declare -g bb__ierrmode="err"

    unset bb__exithooks
    declare -g -A bb__exithooks=()

    _bb_traps_setup

    # configure default message to stderr (with stack trace if >= crit)
    # shellcheck disable=SC2119
    bb_msg_clear_targets
    bb_msg_set_target default_info_err stderr info:err prog notrace
    bb_msg_set_target default_crit_emerg stderr crit:emerg prog trace

    # shellcheck disable=SC2034
    declare -g BASH_BASE_IS_LOADED=0.6.8
}

_bb_main
